// agi.cc
//	Parses general flags on the command line and launches the kernel.
// Usage: nachos -u -z -d <debugflags> ...
//   -u prints entire set of legal flags
//   -z prints copyright string
//   -d causes certain debugging messages to be printed (cf. debug.h)
//
//  NOTE: Other flags are defined for each assignment, and
//  incorrect flag usage is not caught.
//
// Copyright (c) 1992-1996 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#define MAIN
#include "copyright.h"
#undef MAIN

#include "agi.h"
#include "testcase.h"
#include "main.h"

#ifdef AUTOGRADING
#include "autograder.h"
#endif

// global variables
KernelType *kernel;
Debug *debug;

char* AutoGraderInterface::DebugFlags;
char* AutoGraderInterface::AGICommand;

#ifdef AUTOGRADING
#define activecase	AutoGrader::currentcase
#else
TestCase* activecase = new TestCase();
#endif

#define HandleEvent(event, default, args...)				\
  if (activecase!= NULL) return activecase->event(##args);		\
  else return default;

void
AutoGraderInterface::ReadyThread(Thread* thread) {
  HandleEvent(ReadyThread, , thread); }
void
AutoGraderInterface::RunningThread(Thread* thread) {
  HandleEvent(RunningThread, , thread); }
void
AutoGraderInterface::FinishingThread(Thread* thread) {
  HandleEvent(FinishingThread, , thread); }
void
AutoGraderInterface::TLBUpdate(int i, TranslationEntry* entry) {
  HandleEvent(TLBUpdate, , i, entry); }
void
AutoGraderInterface::PageFault(unsigned vaddr) {
  HandleEvent(PageFault, , vaddr); }
void
AutoGraderInterface::Tick(long long time) {
  HandleEvent(Tick,, time); }
void
AutoGraderInterface::Interrupt(int kind, long long time) {
  HandleEvent(Interrupt, , kind, time); }
void
AutoGraderInterface::ConfirmTimerConfiguration(bool random) {
  HandleEvent(ConfirmTimerConfiguration, , random); }
bool
AutoGraderInterface::ExceptionHandler(int which) {
  HandleEvent(ExceptionHandler, TRUE, which); }
void
AutoGraderInterface::ConfirmNetworkReliability(double reliability) {
  HandleEvent(ConfirmNetworkReliability, , reliability); }

//----------------------------------------------------------------------
// AutoGraderInterface::Init
// 	Bootstrap the operating system kernel.  
//	
//	Initialize kernel data structures
//	Call selftest procedure
//	Run the kernel
//
//	"argc" is the number of command line arguments
//	"argv" is an array of strings, one for each command line argument
//----------------------------------------------------------------------
void
AutoGraderInterface::Init(int argc, char** argv)
{
  // install Ctrl+C handler
  CallOnUserAbort(Cleanup);

  ParseCommandLine(argc, argv);

  debug = new Debug(DebugFlags);

  // instantiate, initialize, test, and run the kernel
  dprintf(dbgThread, "Entering kernel");

  kernel = new KernelType(argc, argv);
  kernel->Initialize();

#ifdef AUTOGRADING
  // if grading, transfer control to the autograder
  AutoGrader::ExecuteTest(AGICommand);
#else
  // else just run self test
  kernel->SelfTest();
#endif
  
  kernel->Run();

  ASSERTNOTREACHED();
}

// AutoGraderInterface::Cleanup
//	Ctrl+C handler

void 
AutoGraderInterface::Cleanup(int s) 
{
  cerr.form("\n\nReceived signal %d; attempting to delete kernel\n", s);
  delete kernel;
  ASSERTNOTREACHED();
}

// AutoGraderInterface::ParseCommandLine
//	self-explanatory

void
AutoGraderInterface::ParseCommandLine(int argc, char** argv)
{
  DebugFlags = "";
  AGICommand = "";
  
  // for each argument
  for (int i=0; i<argc; i++) {
    // if it's a switch...
    if (argv[i][0] == '-' && argv[i][1] != 0 && argv[i][2] == 0) {
      // check if it pertains to us
      switch (argv[i++][1]) {
	// debug flags
      case 'd':
	ASSERT(i < argc);
	DebugFlags = argv[i];
	break;
	
	// AGI command
      case 'a':
	ASSERT(i < argc);
	AGICommand = argv[i];
	break;

	// usage
      case 'u':
	cout << "Partial usage: nachos [-z -d debugFlags]\n";
	break;

	// copyright
      case 'z':
	cout << copyright;
	break;

      default:
	continue;
      }
    }
  }
}

int
main(int argc, char **argv)
{
  // get rid of the program name (i hate that)
  argc--; argv++;

  // let the AGI deal with the rest
  AutoGraderInterface::Init(argc, argv);

  ASSERTNOTREACHED();

  return 0;
}
